from typing import Optional

from .resolver import Resolver
from .scope_context import ScopeContext
from .scope_manager import scope_manager
from .scope_type import parse_scope_type, RootType


def ensure_scope(
    scope_id: str,
    scope_type: str,
    parent_id: str = 'root',
    context: Optional[dict[str, str]] = None
) -> None:
    scope_manager.ensure_scope(
        scope_id=scope_id,
        scope_type=parse_scope_type(scope_type),
        parent_id=parent_id,
        context=ScopeContext(context) if context is not None else None
    )


def ensure_root_scope(
    context: Optional[dict[str, str]] = None
) -> None:
    scope_manager.ensure_scope(
        scope_id='root',
        scope_type=RootType(),
        parent_id=None,
        context=ScopeContext(context) if context is not None else None
    )


def get_resolver(scope_id: str = 'root') -> Resolver:
    return scope_manager.get_resolver(scope_id)


def clear_scope(scope_id: str) -> None:
    scope_manager.clear_scope(scope_id=scope_id)


def reset() -> None:
    scope_manager.reset()



from dataclasses import dataclass


@dataclass(frozen=True)
class Condition:
    key: str
    one_of_values: set[str]

    def is_fulfilled_by(self, value_map: dict[str, str]) -> bool:
        if self.key not in value_map:
            return True

        return value_map[self.key] in self.one_of_values


@dataclass(frozen=True)
class Conditions:
    conditions: set[Condition]

    def is_fulfilled_by(self, value_map: dict[str, str]) -> bool:
        return all(condition.is_fulfilled_by(value_map) for condition in self.conditions)


def parse_conditions(**kwargs) -> Conditions:
    return Conditions(
            conditions={
                Condition(
                    key=key,
                    one_of_values=frozenset(value) if isinstance(value, set) else frozenset({value})  # type: ignore
                )
                for key, value in kwargs.items()
            }
        )



from unittest import TestCase

from test_tidipy import context_composition
from test_tidipy.context_composition import StringGenerator, HelloGenerator, TimberGenerator, App
from tidipy import scan, get_resolver, reset, ensure_scope, ensure_root_scope


class TestContext(TestCase):
    def setUp(self) -> None:
        scan(context_composition)

    def tearDown(self) -> None:
        reset()

    def test_production(self):
        ensure_root_scope(context={'environment': 'prod'})
        result = get_resolver()(StringGenerator)

        self.assertEqual(result.generate(), 'Timber')

    def test_specific_type(self):
        ensure_root_scope(context={'environment': 'test'})
        result = get_resolver()(HelloGenerator)

        self.assertEqual(result.generate(), 'Hello')

    def test_resolves_with_additional_values(self):
        ensure_root_scope(context={'environment': 'test', 'extra_key': 'iets'})
        result = get_resolver()(HelloGenerator)

        self.assertEqual(result.generate(), 'Hello')

    def test_cannot_resolve_with_multiple_candidates(self):
        with self.assertRaises(Exception):
            get_resolver()(StringGenerator)

    def test_can_resolve_specific(self):
        result = get_resolver()(TimberGenerator)

        self.assertEqual(result.generate(), 'Timber')

    def test_resolve_app(self):
        ensure_root_scope(context={'environment': 'test'})
        app = get_resolver()(App)

        self.assertEqual(app.generate(), 'Hello')

    def test_cannot_change_context(self):
        ensure_root_scope(context={'environment': 'test'})

        with self.assertRaises(Exception):
            ensure_root_scope(context={'environment': 'prod'})

    def test_cannot_add_to_context(self):
        ensure_root_scope(context={'environment': 'test'})

        with self.assertRaises(Exception):
            ensure_root_scope(context={'new': 'iets'})

    def test_cannot_change_empty_context(self):
        ensure_root_scope()

        with self.assertRaises(Exception):
            ensure_root_scope(context={'a': 'b'})

    def test_cannot_change_context_in_child(self):
        ensure_root_scope(context={'environment': 'test'})

        with self.assertRaises(Exception):
            ensure_scope('child', scope_type='child', context={'environment': 'prod'})

    def test_adding_context_to_child_scope_does_not_influence_root(self):
        ensure_scope(scope_id='child', scope_type='child', context={'environment': 'test'})

        root_resolver = get_resolver()
        child_resolver = get_resolver('child')

        with self.assertRaises(Exception):
            root_resolver(StringGenerator)
        with self.assertRaises(Exception):
            child_resolver(StringGenerator)



from __future__ import annotations

from typing import Optional, Type

from .resolver import Resolver
from .dependency import Dependency, ConcreteDependency, Composer


class DependencyBag:
    def __init__(self, dependencies: dict[str, Dependency]):
        self._dependencies = dependencies

    @staticmethod
    def empty() -> DependencyBag:
        return DependencyBag({})

    def add(self, dependency: Dependency) -> DependencyBag:
        return DependencyBag(
            {
                **self._dependencies,
                dependency.get_id(): dependency
            }
        )

    def remove(self, filter_values: dict[str, str]) -> DependencyBag:
        return DependencyBag(
            {
                dependency_id: dependency
                for dependency_id, dependency in self._dependencies.items()
                if dependency.get_conditions().is_fulfilled_by(filter_values)
            }
        )

    def _get_from_dependency(self, dependency: Dependency, resolver: Resolver):
        concrete = dependency.make_concrete(resolver)

        if dependency.supports_storing():
            self._dependencies[dependency.get_id()] = concrete

        return concrete.value

    def find(
        self,
        dependency_type: Type,
        dependency_id: Optional[str],
        resolver: Resolver
    ) -> Optional[Dependency]:
        candidates = [
            dependency
            for dependency in self._dependencies.values()
            if issubclass(dependency.get_dependency_type(), dependency_type)
        ]
        if dependency_id is not None:
            candidates = [dependency for dependency in candidates if dependency.get_id() == dependency_id]
        if len(candidates) == 1:
            dependency = next(iter(candidates))
            return self._get_from_dependency(dependency, resolver)
        if len(candidates) > 1:
            raise Exception(f'More than 1 candidate for type {dependency_type}')

        return None



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Type, Any, Callable

from .conditions import Conditions
from .resolver import Resolver
from .scope_type import ScopeType


class Dependency(ABC):
    @abstractmethod
    def get_id(self) -> str:
        pass

    @abstractmethod
    def get_dependency_type(self) -> Type:
        pass

    @abstractmethod
    def get_conditions(self) -> Conditions:
        pass

    @abstractmethod
    def make_concrete(self, resolver: Resolver) -> ConcreteDependency:
        pass

    @abstractmethod
    def supports_storing(self) -> bool:
        pass


@dataclass(frozen=True)
class ConcreteDependency(Dependency):
    id: str
    value: Any
    conditions: Conditions

    def get_id(self) -> str:
        return self.id

    def get_dependency_type(self) -> Type:
        return type(self.value)

    def get_conditions(self) -> Conditions:
        return self.conditions

    def make_concrete(self, resolver: Resolver) -> ConcreteDependency:
        return self

    def supports_storing(self) -> bool:
        return True


@dataclass(frozen=True)
class Composer(Dependency):
    id: str
    scope_type: ScopeType
    conditions: Conditions
    factory: Callable[[Resolver], Any]
    dependency_type: Type

    def get_id(self) -> str:
        return self.id

    def __hash__(self) -> int:
        return hash(self.id)

    def get_dependency_type(self) -> Type:
        return self.dependency_type

    def get_conditions(self) -> Conditions:
        return self.conditions

    def make_concrete(self, resolver: Resolver) -> ConcreteDependency:
        return ConcreteDependency(
            id=self.id,
            conditions=self.conditions,
            value=self.factory(resolver)
        )

    def supports_storing(self) -> bool:
        return self.scope_type.supports_storing()
